# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Utility functions for serializing results from autogen."""

from dataclasses import asdict
from typing import Any

from autogen import ChatResult  # type: ignore[import-untyped]


def serialize_results(
    results: ChatResult | list[ChatResult] | dict[int, ChatResult],
) -> dict[str, Any] | list[dict[str, Any]] | dict[int, dict[str, Any]]:
    """Make results serializable.

    Parameters
    ----------
    results : ChatResult | list[ChatResult] | dict[int, ChatResult]
        The results to serialize.

    Returns
    -------
    dict[str, Any] | list[dict[str, Any]] | dict[int, dict[str, Any]]
        The serializable results.
    """
    if isinstance(results, list):
        return [asdict(r) for r in results]
    if isinstance(results, dict):
        return {k: asdict(v) for k, v in results.items()}
    return asdict(results)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Run a synchronous function in an asynchronous way."""

import asyncio
from functools import partial, wraps
from typing import Any, Callable, Coroutine, TypeVar

try:
    from typing import ParamSpec  # type: ignore[unused-ignore,assignment]
except ImportError:
    from typing_extensions import (  # type: ignore[unused-ignore,assignment]
        ParamSpec,
    )

P = ParamSpec("P")
R = TypeVar("R")


def sync_to_async(func: Callable[P, R]) -> Callable[P, Coroutine[Any, Any, R]]:
    """Run a synchronous function in an asynchronous way.

    Parameters
    ----------
    func : Callable
        The synchronous function to run asynchronously.

    Returns
    -------
    Callable
        The asynchronous function.
    """

    @wraps(func)
    async def run_in_executor(*args: P.args, **kwargs: P.kwargs) -> R:
        """Run the synchronous function in an executor.

        Parameters
        ----------
        *args : Any
            The positional arguments for the function.
        **kwargs : Any
            The keyword arguments for the function.

        Returns
        -------
        R
            The result of the synchronous function executed asynchronously.
        """
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            # No event loop running, execute synchronously
            return func(*args, **kwargs)

        partial_func = partial(func, *args, **kwargs)
        return await loop.run_in_executor(None, partial_func)

    return run_in_executor



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Callback enabled IO stream for Waldiez Studio."""

from typing import Any, Callable

from autogen.events import BaseEvent  # type: ignore
from autogen.io import IOStream  # type: ignore


class DelegatedIOStream(IOStream):
    """Delegate the input and output operations to provided callbacks."""

    def __init__(
        self,
        on_input: Callable[..., str],
        on_output: Callable[..., None],
        is_async: bool = False,
        on_send: Callable[[BaseEvent], None] | None = None,
    ) -> None:
        super().__init__()
        self._on_input = on_input
        self._on_output = on_output
        self._on_send = on_send
        self._is_async = is_async

    def print(self, *args: Any, **kwargs: Any) -> None:
        """Prints to the output stream using the provided callback.

        Parameters
        ----------
        *args : Any
            Positional arguments to pass to the output callback.
        **kwargs : Any
            Keyword arguments to pass to the output callback.
        """
        self._on_output(*args, **kwargs)

    def input(self, prompt: str = "", *, password: bool = False) -> str:
        """Structured input from stdin.

        Parameters
        ----------
        prompt : str, optional
            The prompt to display. Defaults to "".
        password : bool, optional
            Whether to read a password. Defaults to False.

        Returns
        -------
        str
            The line read from the input stream.
        """
        # let's keep an eye here:
        # https://github.com/ag2ai/ag2/blob/main/autogen/agentchat/conversable_agent.py#L2973
        # reply = await iostream.input(prompt) ???? (await???)
        if self._is_async:
            # let's make a coroutine to just return the input method
            async def async_input() -> str:
                """Asynchronous input method.

                Returns
                -------
                str
                    The line read from the input stream.
                """
                return self._on_input(prompt=prompt, password=password)

            return async_input()  # type: ignore
        return self._on_input(prompt=prompt, password=password)

    def send(self, message: BaseEvent) -> None:
        """Sends a message to the input stream.

        Parameters
        ----------
        message : BaseEvent
            The message to send.
        """
        if self._on_send:
            self._on_send(message)
        else:
            self.print(
                message.model_dump_json(),
                flush=True,
                end="",
            )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Module for defining task states using an enumeration."""

from enum import Enum, auto


class TaskState(Enum):
    """Enumeration for task states."""

    NOT_STARTED = auto()
    RUNNING = auto()
    COMPLETED = auto()



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=too-many-try-statements,broad-exception-caught
"""TaskRunner class to manage task execution and WebSocket communication."""

import asyncio
import logging
import os
import uuid
from typing import Any, Callable
from urllib.parse import quote

from autogen import ChatResult  # type: ignore
from autogen.io import IOStream  # type: ignore
from fastapi import WebSocket, WebSocketDisconnect
from waldiez import WaldiezRunner
from waldiez.io import UserResponse

from .delegated_iostream import DelegatedIOStream
from .paths import get_root_dir, id_to_path
from .restart import restart_process
from .results import serialize_results
from .task_state import TaskState

LOG = logging.getLogger(__name__)
Data = str | bytes

MAX_ACTIVE_TASKS = 10
active_tasks_semaphore = asyncio.Semaphore(MAX_ACTIVE_TASKS)


# pylint: disable=too-many-instance-attributes
class TaskRunner:
    """TaskRunner to manage task execution and WebSocket communication."""

    task_state: TaskState

    def __init__(
        self, task_id: str, websocket: WebSocket, input_timeout: float = 30
    ) -> None:
        self.websocket = websocket
        self.task_id = task_id
        self.task_state = TaskState.NOT_STARTED
        self.running_task: asyncio.Task[None] | None = None
        self.comm_handler_task: asyncio.Task[None] | None = None
        self.loop = asyncio.get_running_loop()
        self.stop_event = asyncio.Event()
        self.input_timeout = input_timeout

        self.cwd = os.getcwd()
        self.latest_request_id: str | None = None
        self.pending_inputs: dict[str, asyncio.Future[str]] = {}
        self.output_messages: asyncio.Queue[str] = asyncio.Queue()

    def _handle_user_response(
        self,
        data: dict[str, Any],
    ) -> None:
        """Handle user response data.

        Parameters
        ----------
        data : dict[str, Any]
            The data received from the user response.
        """
        try:
            response = UserResponse.model_validate(data)
            request_id = response.request_id
            self.latest_request_id = request_id
            user_input = response.to_string(
                uploads_root=id_to_path(self.task_id).parent,
                base_name=request_id,
            )
            # LOG.debug("Input response received: %s", user_input)
            future = self.pending_inputs.get(request_id)
            if future and not future.done():  # pragma: no branch
                self.pending_inputs.pop(request_id, None)
                future.set_result(user_input)
        except Exception as exc:
            LOG.error("Validation error for user response: %s", exc)

    async def listen(self) -> None:
        """Listen for WebSocket messages and handle actions.

        Raises
        ------
        asyncio.CancelledError
            If the task is cancelled.
        """
        try:
            while True:
                try:
                    data: dict[str, Any] = await self.websocket.receive_json()
                    # LOG.debug("Received data: %s", data)
                    action = data.get("action")

                    if action:
                        await self._handle_action(action)
                    elif data.get("type") == "input_response":
                        self._handle_user_response(data)
                except WebSocketDisconnect:
                    LOG.info("WebSocket disconnected.")
                    break
        except asyncio.CancelledError as exc:
            LOG.info("WebSocket listen task cancelled: %s", exc)
            raise
        except BaseException as err:
            LOG.error("WebSocket error: %s", err)
            await self.websocket.close(code=1008, reason="Connection error.")

    async def stop_task(self) -> None:
        """Stop the currently running task and clean up resources."""
        self.stop_event.set()

        if (
            self.running_task and not self.running_task.done()
        ):  # pragma: no branch
            self.running_task.cancel()
            try:
                await asyncio.wait_for(self.running_task, timeout=5)
            except asyncio.TimeoutError:
                LOG.warning("Forcefully stopping running task due to timeout.")
            except asyncio.CancelledError:
                pass

        if (
            self.comm_handler_task and not self.comm_handler_task.done()
        ):  # pragma: no branch
            self.comm_handler_task.cancel()
            try:
                await asyncio.wait_for(self.comm_handler_task, timeout=2)
            except asyncio.TimeoutError:
                LOG.warning(
                    "Forcefully stopping comm handler task due to timeout."
                )
            except asyncio.CancelledError:
                pass

        self.task_state = TaskState.COMPLETED
        await self.send_message("info", "Task stopped.")
        # Restart the process to ensure a clean state
        self._restart_process()

    def _restart_process(self) -> None:  # pragma: no cover
        """Restart the current process to ensure a clean state."""
        is_testing = (
            os.environ.get("WALDIEZ_STUDIO_TESTING", "False").lower() == "true"
        )
        if is_testing:
            return
        os.chdir(self.cwd)
        restart_process()

    async def run(self) -> None:
        """Run the task in a separate thread and handle communication.

        Raises
        ------
        asyncio.CancelledError
            If the task is cancelled.
        BaseException
            If there is an error during task execution.
        """
        async with active_tasks_semaphore:
            if self.task_state == TaskState.RUNNING:
                await self.send_message("info", "Task is already running.")
                return
            if self.task_state == TaskState.COMPLETED:
                await self.send_message("info", "Task is already completed.")
                return

            self.task_state = TaskState.RUNNING
            self.stop_event.clear()

            # Start comm handler to forward outputs
            self.comm_handler_task = asyncio.create_task(
                self._handle_thread_communication()
            )

            try:
                self.running_task = asyncio.create_task(self._run())
                await self.running_task
            except (asyncio.CancelledError, BaseException) as exc:
                LOG.error("Task run error: %s", exc)
                raise
            finally:
                self.stop_event.set()
                if self.comm_handler_task and not self.comm_handler_task.done():
                    self.comm_handler_task.cancel()
                    try:
                        await self.comm_handler_task
                    except asyncio.CancelledError:
                        pass
                self.running_task = None
                # pylint: disable=redefined-variable-type
                self.task_state = TaskState.COMPLETED

    async def _handle_action(self, action: str) -> None:
        """Handle actions received from the WebSocket.

        Parameters
        ----------
        action : str
            The action to handle (e.g., "start", "stop", "status").
        """
        if action == "start":
            await self._handle_start()
        elif action == "stop":
            await self._handle_stop()
        elif action == "status":
            await self.send_message("status", self.task_state.name)

    async def _handle_start(self) -> None:
        """Handle the start action for the task."""
        if self.task_state in (TaskState.NOT_STARTED, TaskState.COMPLETED):
            self.task_state = TaskState.NOT_STARTED
            asyncio.create_task(self.run())
        elif self.task_state == TaskState.RUNNING:  # pragma: no branch
            await self.send_message("info", "Task is already running.")

    async def _handle_stop(self) -> None:
        """Handle the stop action for the task."""
        if self.task_state == TaskState.RUNNING:
            await self.stop_task()
        else:
            await self.send_message("info", "No running task to stop.")

    async def _run(self) -> None:
        on_input, on_output = self._create_thread_safe_callbacks()

        results = await asyncio.to_thread(
            run_task,
            self.task_id,
            on_input,
            on_output,
        )

        if not results:  # pragma: no cover
            raise ValueError("No results returned.")

        await self.websocket.send_json(
            {
                "type": "results",
                "data": serialize_results(results),
            }
        )

    async def _handle_thread_communication(self) -> None:
        try:
            while not self.stop_event.is_set():
                try:
                    # Use asyncio.wait_for to avoid infinite blocking
                    msg = await asyncio.wait_for(
                        self.output_messages.get(), timeout=0.1
                    )
                    await self._send_output(msg)
                except asyncio.TimeoutError:
                    continue
        except asyncio.CancelledError:  # pragma: no cover
            LOG.warning("Thread communication handler cancelled.")

    def _create_thread_safe_callbacks(
        self,
    ) -> tuple[Callable[[str], str], Callable[..., None]]:
        # pylint: disable=unused-argument
        def on_input(prompt: str, *, password: bool = False) -> str:
            """Thread-safe input callback.

            Parameters
            ----------
            prompt : str
                The input prompt to display.
            password : bool, optional
                Whether the input is a password (default is False).

            Returns
            -------
            str
                The input received from the user.
            """
            coro = self._send_input_request(prompt)
            future = asyncio.run_coroutine_threadsafe(coro, self.loop)
            return future.result(timeout=self.input_timeout)

        def on_output(
            *objects: Any, sep: str = " ", end: str = "\n", flush: bool = False
        ) -> None:
            """Thread-safe output callback.

            Parameters
            ----------
            objects : Any
                The objects to output.
            sep : str, optional
                The separator between objects (default is a space).
            end : str, optional
                The string appended after the last object.
                Defaults to a new line (\\n).
            flush : bool, optional
                Whether to flush the output (default is False).
            """
            message = sep.join(str(obj) for obj in objects) + end
            self.output_messages.put_nowait(message)

        return on_input, on_output

    async def _send_input_request(self, prompt: str) -> str:
        request_id = uuid.uuid4().hex
        self.latest_request_id = request_id
        future = self.loop.create_future()
        self.pending_inputs[request_id] = future

        await self.websocket.send_json(
            {
                "id": request_id,
                "type": "input_request",
                "request_id": request_id,
                "prompt": prompt,
                "password": False,
                "content": [{"type": "text", "text": prompt}],
            }
        )

        try:
            return await asyncio.wait_for(future, timeout=self.input_timeout)
        except asyncio.TimeoutError:
            LOG.warning(
                "Input request timed out after %.1f seconds.",
                self.input_timeout,
            )
            return ""
        finally:
            self.pending_inputs.pop(request_id, None)

    def _replace_image_placeholders(self, message: str) -> str:
        """Replace image placeholders in the message with actual image links.

        Parameters
        ----------
        message : str
            The message containing image placeholders.

        Returns
        -------
        str
            The message with image placeholders replaced by actual links.
        """
        if (
            "<image>" not in message or not self.latest_request_id
        ):  # pragma: no cover
            return message
        uploads_root = id_to_path(self.task_id).parent
        public_image_path = str(uploads_root.relative_to(get_root_dir()))
        if public_image_path.startswith("."):  # pragma: no branch
            public_image_path = public_image_path[1:]
        if (
            not public_image_path or public_image_path == "."
        ):  # pragma: no branch
            public_image_path = ""
        else:  # pragma: no cover
            public_image_path += "/"

        public_image_path = quote(public_image_path, safe="/")
        image_path = (
            f"/api/workspace/download?path={public_image_path}"
            f"{self.latest_request_id}.png"
        )
        return message.replace("<image>", str(image_path))

    async def _send_output(self, message: str) -> None:
        processed_message = self._replace_image_placeholders(message)
        await self.websocket.send_json(
            {
                "type": "print",
                "data": processed_message,
            }
        )

    async def send_message(self, msg_type: str, data: Data) -> None:
        """Send a message through the WebSocket.

        Parameters
        ----------
        msg_type : str
            The type of the message (e.g., "status", "info", "error").
        data : Data
            The data to send, which can be a string or bytes.
        """
        await self.websocket.send_json(
            {
                "type": msg_type,
                "data": data,
            }
        )


def run_task(
    task_id: str,
    on_input: Callable[..., str],
    on_output: Callable[..., None],
) -> ChatResult | list[ChatResult] | dict[int, ChatResult]:
    """Run a Waldiez task with the given task ID.

    Parameters
    ----------
    task_id : str
        The ID of the task to run.
    on_input : Callable[..., str]
        Callback function to handle input requests.
    on_output : Callable[..., None]
        Callback function to handle output messages.

    Returns
    -------
    ChatResult | list[ChatResult] | dict[int, ChatResult]
        The results of the task execution.

    Raises
    ------
    Exception
        If there is an error loading the task file or running the task.
    """
    file_path = id_to_path(task_id)
    output_path = file_path.with_suffix(".py")

    try:
        runner = WaldiezRunner.load(file_path)
    except Exception as exc:
        LOG.error("Error loading task file: %s", exc)
        raise RuntimeError(
            f"Failed to load task file {file_path}: {exc}"
        ) from exc

    try:
        stream = DelegatedIOStream(
            on_input=on_input,
            on_output=on_output,
            is_async=runner.is_async,
        )
        with IOStream.set_default(stream):
            return runner.run(
                output_path=output_path, uploads_root=file_path.parent
            )
    except Exception as exc:
        raise RuntimeError(f"Failed to run task {task_id}: {exc}") from exc



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Common utilities for the api routes."""

import logging
import re
from pathlib import Path
from typing import Optional
from urllib.parse import unquote

from fastapi import Depends, HTTPException, Query

from waldiez_studio.utils.paths import get_root_dir

LOG = logging.getLogger(__name__)


def get_root_directory() -> Path:
    """Get the root directory of the workspace.

    Returns
    -------
    Path
        The root directory of the workspace.
    """
    return get_root_dir()


def _resolve_path(path: Path, strict: bool = False) -> Path:
    """Resolve a path and ensure it exists.

    Parameters
    ----------
    path : Path
        The path to resolve.
    strict : bool, optional
        Whether to strictly enforce the path, by default False.

    Returns
    -------
    Path
        The resolved path.

    Raises
    ------
    FileNotFoundError
        If the path does not exist.
    """
    try:
        resolved_path = path.resolve(strict=strict)
    except FileNotFoundError as error:
        raise FileNotFoundError("Error: Path not resolved") from error
    except BaseException as error:
        raise ValueError("Error: Path not resolved") from error
    return resolved_path


def sanitize_path(root_dir: Path, path: str, strict: bool = False) -> Path:
    """Sanitize a path and ensure it is within the root directory.

    Parameters
    ----------
    root_dir : Path
        The root directory to check against.
    path : str
        The path to sanitize.
    strict : bool, optional
        Whether to strictly enforce the path, by default False.

    Returns
    -------
    Path
        The sanitized path.

    Raises
    ------
    ValueError
        If the path is invalid or traverses outside the root directory.
    BaseException
        If the path cannot be resolved.
    """
    if path in ("/", ""):
        return root_dir
    try:
        unquoted_path = unquote(path, errors="strict")
        if not unquoted_path.strip():
            return root_dir
    except BaseException as error:
        raise ValueError("Error: Invalid path") from error
    if unquoted_path in ("/", ""):
        return root_dir
    # Reject paths with invalid characters
    if re.search(r"[<>:\"|?*]", unquoted_path):
        raise ValueError("Error: Invalid path")
    safe_path = root_dir / Path(unquoted_path.strip("/"))
    try:
        resolved_path = _resolve_path(safe_path, strict=strict)
    except BaseException as error:
        raise error
    if not resolved_path.is_relative_to(root_dir):
        raise ValueError("Error: Invalid path: Path is outside root directory")
    return resolved_path


def check_path(
    path: str,
    root_dir: Path,
    path_type: str = "Path",
    must_exist: bool = True,
    must_not_exist: bool = False,
    must_be_dir: bool = False,
    must_be_file: bool = False,
    must_have_extension: Optional[str] = None,
) -> Path:
    """Reusable dependency for sanitizing paths.

    Parameters
    ----------
    path : str
        The path to sanitize.
    root_dir : Path
        The root directory to use.
    path_type : str, optional
        The type of path, by default "Path" ("Path", "File" or "Directory").
    must_exist : bool, optional
        Whether the path must exist, by default True.
    must_not_exist : bool, optional
        Whether the path must not exist, by default False.
    must_be_dir : bool, optional
        Whether the path must be a directory, by default False.
    must_be_file : bool, optional
        Whether the path must be a file, by default False.
    must_have_extension : Optional[str], optional
        The extension the path must have, by default None.

    Returns
    -------
    Path
        The sanitized path.

    Raises
    ------
    HTTPException
        If the path is invalid.
    """
    thing = "File" if must_be_file else "Directory" if must_be_dir else "Path"
    try:
        the_path = sanitize_path(root_dir, path, must_exist)
    except FileNotFoundError as error:
        LOG.warning("Invalid path: %s, %s, %s", path, root_dir, error)
        raise HTTPException(
            status_code=404, detail=f"Error: {thing} not found"
        ) from error
    except ValueError as error:
        LOG.warning("Invalid path: %s, %s, %s", path, root_dir, error)
        raise HTTPException(
            status_code=400, detail=f"Error: Invalid {path_type.lower()}"
        ) from error
    thing = (
        _guess_path_type(must_be_dir, must_be_file, the_path)
        if path_type == "Path"
        else path_type
    )
    if must_exist and not the_path.exists():
        raise HTTPException(status_code=404, detail=f"Error: {thing} not found")
    if must_not_exist and the_path.exists():
        raise HTTPException(
            status_code=400, detail=f"Error: {thing} already exists"
        )
    if must_be_dir and not the_path.is_dir():
        raise HTTPException(status_code=400, detail="Error: Not a directory")
    if must_be_file and not the_path.is_file():
        raise HTTPException(status_code=400, detail="Error: Not a file")
    if must_have_extension and the_path.suffix != must_have_extension:
        raise HTTPException(status_code=400, detail="Error: Invalid file type")
    return the_path.resolve(must_exist)


def _guess_path_type(
    should_be_dir: bool, should_be_file: bool, path: Path
) -> str:
    """Get the path type for error messages.

    Parameters
    ----------
    should_be_dir : bool
        Whether the path is expected to be a directory.
    should_be_file : bool
        Whether the path is expected to be a file.
    path : Path
        The path to check.

    Returns
    -------
    str
        The path type.
    """
    if should_be_dir:
        return "Directory"
    if should_be_file:
        return "File"
    if path.is_dir():
        return "Directory"
    if path.is_file():
        return "File"
    return "Path"


def get_new_folder_name(root_dir: Path, folder_name: str) -> str:
    """Get a new folder name that does not already exist.

    Parameters
    ----------
    root_dir : Path
        The root directory to check for existing folders.
    folder_name : str
        The original folder name.

    Returns
    -------
    str
        The new folder name.
    """
    # add (n) to the folder name if it already exists
    n = 0
    new_folder_name = folder_name
    while (root_dir / new_folder_name).exists():
        n += 1
        new_folder_name = f"{folder_name} ({n})"
    return new_folder_name


def get_new_file_name(root_dir: Path, file_name: str) -> str:
    """Get a new file name that does not already exist.

    Parameters
    ----------
    root_dir : Path
        The root directory to check for existing files.
    file_name : str
        The original file name.

    Returns
    -------
    str
        The new file name.
    """
    # add (n) to the file name if it already exists
    n = 0
    new_file_name = file_name
    while (root_dir / new_file_name).exists():
        n += 1
        parts = file_name.rsplit(".", maxsplit=1)
        if len(parts) == 2:
            new_file_name = f"{parts[0]} ({n}).{parts[1]}"
        else:
            new_file_name = f"{file_name} ({n})"
    return new_file_name


def check_flow_path(
    path: str = Query(..., description="The path to the flow."),
    root_dir: Path = Depends(get_root_directory),
) -> Path:
    """Validate the flow path.

    Parameters
    ----------
    path : str
        The path to the flow.
    root_dir : Path
        The root directory of the workspace.

    Returns
    -------
    Path
        The validated flow path relative to the root directory.
    Raises
    ------
    HTTPException
        If the path is invalid, the file does not exist,
        the file is not a file, or the file type is invalid.
    """
    return check_path(
        path,
        root_dir,
        path_type="File",
        must_exist=True,
        must_be_file=True,
        must_be_dir=False,
        must_have_extension=".waldiez",
    )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Restart the Waldiez Studio process."""

import logging
import os
import shlex
import sys

LOG = logging.getLogger(__name__)


def restart_process() -> None:  # pragma: no cover
    """Restart the current process with the same command line arguments."""
    to_call = [
        sys.executable,
        "-m",
        "waldiez_studio",
    ] + skip_positional_prefix(sys.argv[1:])
    to_log = " ".join(shlex.quote(arg) for arg in to_call)
    LOG.info("Restarting process with command: %s", to_log)
    os.execv(sys.executable, to_call)  # nosemgrep # nosec


def skip_positional_prefix(argv: list[str]) -> list[str]:
    """Skip all leading args until the first option/flag (starts with '-').

    Parameters
    ----------
    argv : list[str]
        The command line arguments.
    Returns
    -------
    list[str]
        The arguments starting from the first option/flag.
    """
    for i, arg in enumerate(argv):
        if arg.startswith("-"):
            return argv[i:]
    return []



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Directories to use for user uploads and flow outputs."""

import os
from functools import lru_cache
from pathlib import Path


@lru_cache
def get_root_dir(user_id: str = "default") -> Path:
    """Get the root directory for flows, user uploads, etc.

    Parameters
    ----------
    user_id : str
        The user ID (if using multi-user mode)

    Returns
    -------
    Path
        The root waldiez directory
    """
    files_root = Path(__file__).parent.parent / "files"
    root_dir = files_root / user_id
    is_testing = (
        os.environ.get("WALDIEZ_STUDIO_TESTING", "False").lower() == "true"
    )
    if is_testing:
        root_dir = (
            Path(os.environ.get("WALDIEZ_STUDIO_ROOT_DIR", str(files_root)))
            / user_id
        )
    root_dir.mkdir(parents=True, exist_ok=True)
    return root_dir.resolve(True)


def get_static_dir() -> Path:
    """Get the static directory.

    Returns
    -------
    Path
        The static directory
    """
    root_dir = Path(__file__).parent.parent / "static"
    is_testing = (
        os.environ.get("WALDIEZ_STUDIO_TESTING", "False").lower() == "true"
    )
    if is_testing:
        root_dir = Path(
            os.environ.get("WALDIEZ_STUDIO_ROOT_DIR", str(root_dir))
        )
    return root_dir


def path_to_id(path: Path) -> str:
    """Convert a Path object to a Base64-encoded string identifier.

    Parameters
    ----------
    path : Path
        The Path object to convert.

    Returns
    -------
    str
        Base64-encoded string identifier.
    """
    return path.as_posix().encode("utf-8").hex()


def id_to_path(identifier: str) -> Path:
    """Convert a Base64-encoded string identifier back to a Path object.

    Parameters
    ----------
    identifier : str
        The Base64-encoded string identifier.

    Returns
    -------
    Path
        Reconstructed Path object.
    """
    return Path(bytes.fromhex(identifier).decode("utf-8"))



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Websocket routes."""

import asyncio
import logging
from pathlib import Path

from fastapi import APIRouter, Depends, WebSocket

from waldiez_studio.utils.paths import path_to_id
from waldiez_studio.utils.task_runner import TaskRunner

from .common import check_path, get_root_directory

router = APIRouter()

LOG = logging.getLogger(__name__)


# pylint: disable=unused-argument
@router.websocket("/ws")
async def websocket_endpoint(
    websocket: WebSocket,
    path: str,
    root_dir: Path = Depends(get_root_directory),
) -> None:
    """Websocket endpoint.

    Parameters
    ----------
    websocket : WebSocket
        The websocket.
    path : str
        The path to the flow.
    root_dir : Path
        The root directory of the workspace
    """
    # pylint: disable=broad-except
    try:
        flow_path = check_path(
            path,
            root_dir,
            path_type="File",
            must_exist=True,
            must_be_file=True,
            must_be_dir=False,
            must_have_extension=".waldiez",
        )
    except BaseException as exc:
        LOG.error("Error: %s", exc)
        await websocket.close(code=1008, reason="Invalid path")
        return
    await websocket.accept()
    task_id = path_to_id(flow_path)
    task_runner = TaskRunner(
        task_id=task_id,
        websocket=websocket,
    )
    # pylint: disable=broad-except
    try:
        await asyncio.gather(
            task_runner.listen(),
        )
    except BaseException as exc:
        LOG.warning("WebSocket connection error: %s", exc)
    finally:
        try:
            await websocket.close(code=1006, reason="Connection error")
        except BaseException:
            pass



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Workspace API routes."""

import logging
import shutil
import tempfile
from pathlib import Path
from typing import List, Optional

import aiofiles
import aiofiles.os
from fastapi import (
    APIRouter,
    BackgroundTasks,
    Depends,
    File,
    Form,
    HTTPException,
    Query,
    UploadFile,
)
from fastapi.responses import FileResponse

from waldiez_studio.models import (
    MessageResponse,
    PathItem,
    PathItemCreateRequest,
    PathItemListResponse,
    PathItemRenameRequest,
    PathItemType,
)
from waldiez_studio.utils.sync import sync_to_async

from .common import (
    check_path,
    get_new_file_name,
    get_new_folder_name,
    get_root_directory,
)

api = APIRouter()
LOG = logging.getLogger(__name__)

CHUNK_SIZE = 1024 * 1024  # 1 MB chunks
MAX_FILE_SIZE = 50 * 1024 * 1024  # 50 MB maximum file size


@api.get("/workspace", response_model=PathItemListResponse)
async def list_items(
    parent: Optional[str] = Query(
        default=None,
        description=(
            "The directory path to list. "
            "If not provided, the root directory will be listed."
        ),
    ),
    root_dir: Path = Depends(get_root_directory),
) -> PathItemListResponse:
    """
    List files and folders in a directory.

    Parameters
    ----------
    parent : Optional[str]
        The directory path to list.
        If not provided, the root directory will be listed.
    root_dir : Path
        The root directory to use.

    Returns
    -------
    PathItemListResponse
        A list of files and folders in the specified directory.

    Raises
    ------
    HTTPException
        If the directory does not exist.
    """
    target_dir = (
        root_dir
        if not parent
        else check_path(parent, root_dir=root_dir, must_be_dir=True)
    )

    def sync_list_items(directory: Path) -> List[PathItem]:
        """List files and folders in a given directory.

        Parameters
        ----------
        directory : Path
            The directory to list.

        Returns
        -------
        List[PathItem]
            A list of files and folders in the specified directory.
        """
        items = [
            PathItem(
                name=item.name,
                path=str(item.relative_to(root_dir)),
                type="folder" if item.is_dir() else "file",
            )
            for item in directory.iterdir()
        ]
        return items

    entries = await sync_to_async(sync_list_items)(target_dir)
    return PathItemListResponse(items=entries)


@api.post(
    "/workspace",
    response_model=PathItem,
    responses={
        "200": {"description": "The newly created file or folder"},
        "400": {"description": "Error: Invalid path"},
        "404": {"description": "Error: Path or parent directory not found"},
    },
)
async def create_file_or_folder(
    data: PathItemCreateRequest,
    root_dir: Path = Depends(get_root_directory),
) -> PathItem:
    """Create a new file or folder in the root directory.

    Parameters
    ----------
    data : PathItemCreateRequest
        The type of the item to create ("file" or "folder")
        and the parent folder path.
    root_dir : Path
        The root directory to use.

    Returns
    -------
    PathItem
        The newly created file or folder.

    Raises
    ------
    HTTPException
        If the file or folder already exists.
    """
    target_dir = (
        root_dir
        if not data.parent
        else check_path(data.parent, root_dir=root_dir, must_be_dir=True)
    )

    new_name = (
        get_new_folder_name(target_dir, "New Folder")
        if data.type == "folder"
        else get_new_file_name(target_dir, "Untitled.waldiez")
    )
    full_path = target_dir / new_name
    if full_path.exists():
        raise HTTPException(
            status_code=400, detail="Error: File or folder already exists"
        )
    return await create_new_path_item(
        full_path, item_type=data.type, root_dir=root_dir
    )


@api.post(
    "/workspace/rename",
    response_model=PathItem,
    responses={
        "200": {"description": "The renamed file or folder"},
        "400": {"description": "Error: Invalid path"},
        "404": {"description": "Error: Source file or folder does not exist"},
        "500": {"description": "Failed to rename file or folder"},
    },
)
async def rename_file_or_folder(
    data: PathItemRenameRequest,
    root_dir: Path = Depends(get_root_directory),
) -> PathItem:
    """
    Rename a file or folder.

    Parameters
    ----------
    data : PathItemRenameRequest
        The old and new paths for the file or folder.
    root_dir : Path
        The root directory to use.

    Returns
    -------
    PathItem
        The renamed file or folder.

    Raises
    ------
    HTTPException
        If the old path does not exist,
        the new path already exists,
        or renaming fails.
    """
    old_path = data.old_path
    new_path = data.new_path
    old_target = check_path(old_path, root_dir=root_dir, must_exist=True)
    new_target = check_path(
        new_path, root_dir=root_dir, must_exist=False, must_not_exist=True
    )
    try:
        old_target.rename(new_target)
    except BaseException as error:
        raise HTTPException(
            status_code=500, detail="Failed to rename file or folder"
        ) from error

    return PathItem(
        name=new_target.name,
        path=str(new_target.relative_to(root_dir)),
        type="folder" if new_target.is_dir() else "file",
    )


@api.post(
    "/workspace/upload",
    response_model=PathItem,
    responses={
        "200": {"description": "The uploaded file"},
        "400": {"description": "Invalid request"},
        "404": {"description": "Error: Directory not found"},
        "500": {"description": "Failed to upload file"},
    },
)
async def upload_file(
    path: str = Form(...),
    file: UploadFile = File(...),
    root_dir: Path = Depends(get_root_directory),
) -> PathItem:
    """Upload a file in chunks.

    Parameters
    ----------
    path : str
        The directory path where the file will be uploaded.
    file : UploadFile
        The file to upload.
    root_dir : Path
        The root directory to use.

    Returns
    -------
    PathItem
        The uploaded file.

    Raises
    ------
    HTTPException
        If the file exceeds the maximum size,
        if the already exists or there is an upload error.
    """
    if not file.filename:
        raise HTTPException(status_code=400, detail="Error: Invalid file name")
    target_dir = check_path(
        path, root_dir=root_dir, must_exist=True, must_be_dir=True
    )
    full_path = target_dir / file.filename
    if full_path.exists():
        raise HTTPException(
            status_code=400, detail="Error: File already exists"
        )
    file_size = 0
    # pylint: disable=too-many-try-statements
    try:
        async with aiofiles.open(full_path, "wb") as f:
            while chunk := await file.read(CHUNK_SIZE):
                file_size += len(chunk)
                if file_size > MAX_FILE_SIZE:
                    raise HTTPException(
                        status_code=400,
                        detail="Error: File exceeds maximum size",
                    )
                await f.write(chunk)
    except HTTPException as error:
        full_path.unlink(missing_ok=True)
        raise error
    except BaseException as error:
        full_path.unlink(missing_ok=True)
        raise HTTPException(
            status_code=500, detail="Failed to upload file"
        ) from error

    return PathItem(
        name=full_path.name,
        path=str(full_path.relative_to(root_dir)),
        type="file",
    )


@api.get(
    "/workspace/download",
    responses={
        "200": {"description": "The file was downloaded"},
        "400": {"description": "Error: Invalid path"},
        "404": {"description": "Error: File or folder not found"},
        "500": {"description": "Failed to download file or folder"},
    },
)
async def download_file_or_folder(
    path: str,
    background_tasks: BackgroundTasks,
    root_dir: Path = Depends(get_root_directory),
) -> FileResponse:
    """Download a file or folder.

    Parameters
    ----------
    path : str
        The path to the file or folder to download.
    background_tasks : BackgroundTasks
        The background tasks to use for deleting temporary files.
    root_dir : Path
        The root directory to use.

    Returns
    -------
    FileResponse
        The file or folder to download.

    Raises
    ------
    HTTPException
        If the file or folder does not exist or the download fails.
    """
    # make a zip if folder
    target_path = check_path(
        path, root_dir=root_dir, must_exist=True, must_be_dir=False
    )
    if target_path.is_file():
        return FileResponse(target_path, filename=target_path.name)
    return await sync_to_async(download_zip_sync)(
        target_path, root_dir, background_tasks
    )


@api.delete(
    "/workspace",
    response_model=MessageResponse,
    responses={
        "200": {"description": "The file or folder was deleted"},
        "400": {"description": "Error: Invalid path"},
        "404": {"description": "Error: File or folder not found"},
        "500": {"description": "Failed to delete file or folder"},
    },
)
async def delete_file_or_folder(
    path: str, root_dir: Path = Depends(get_root_directory)
) -> MessageResponse:
    """Delete a file or folder.

    Parameters
    ----------
    path : str
        The path to the file or folder to delete.
    root_dir : Path
        The root directory to use.

    Returns
    -------
    MessageResponse
        A success message.

    Raises
    ------
    HTTPException
        If the file or folder does not exist.
    """
    target_path = check_path(
        path, root_dir=root_dir, must_exist=True, must_be_dir=False
    )
    thing = "Folder" if target_path.is_dir() else "File"
    try:
        if target_path.is_dir():
            rmtree = sync_to_async(shutil.rmtree)
            # pylint: disable=line-too-long
            # fmt: off
            await rmtree(target_path, ignore_errors=True)  # type: ignore[unused-ignore,call-arg]  # noqa: E501
            # fmt: on
        else:
            await aiofiles.os.remove(target_path)
    except BaseException as error:
        raise HTTPException(
            status_code=500, detail=f"Failed to delete {thing.lower()}"
        ) from error

    return MessageResponse(message=f"{thing} deleted successfully")


async def create_new_path_item(
    full_path: Path, item_type: PathItemType, root_dir: Path
) -> PathItem:
    """Create a new file or folder.

    Parameters
    ----------
    full_path : Path
        The full path where the file or folder will be created.
    item_type : PathItemType
        The type of the item to create ("file" or "folder").
    root_dir : Path
        The root directory to use.

    Returns
    -------
    PathItem
        The newly created file or folder.
    Raises
    ------
    HTTPException
        If the item type is invalid or creation fails.
    """
    try:
        if item_type == "folder":
            full_path.mkdir(exist_ok=False)
        elif item_type == "file":
            async with aiofiles.open(full_path, mode="w") as f:
                await f.write('{"type": "flow"}')
    except BaseException as error:
        raise HTTPException(
            status_code=500, detail=f"Error: Failed to create {item_type}"
        ) from error
    return PathItem(
        name=full_path.name,
        path=str(full_path.relative_to(root_dir)),
        type=item_type,
    )


def download_zip_sync(
    target_path: Path, root_dir: Path, background_tasks: BackgroundTasks
) -> FileResponse:
    """Download a folder as a zip file.

    Parameters
    ----------
    target_path : Path
        The path to the folder to download.
    root_dir : Path
        The root directory to use.
    background_tasks : BackgroundTasks
        The background tasks to use for deleting temporary files.

    Returns
    -------
    FileResponse
        The zip file to download.

    Raises
    ------
    HTTPException
        If the download (archiving) fails.
    """
    # pylint: disable=too-many-try-statements
    try:
        # Create a temporary directory
        temp_dir = tempfile.mkdtemp()
        temp_dir_path = Path(temp_dir)
        zip_path = temp_dir_path / f"{target_path.name}.zip"

        # Create the archive
        archive = shutil.make_archive(
            base_name=str(zip_path.with_suffix("")),
            format="zip",
            root_dir=str(root_dir),
            base_dir=target_path.name,
        )

        # Add cleanup task for the temporary directory
        def cleanup() -> None:
            """Clean up the temporary directory."""
            shutil.rmtree(temp_dir, ignore_errors=True)

        background_tasks.add_task(cleanup)

        # Return the archive as a FileResponse
        return FileResponse(
            archive, filename=zip_path.name, background=background_tasks
        )

    except BaseException as error:
        raise HTTPException(
            status_code=500, detail="Error: Failed to download folder"
        ) from error
